/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * ---------------------------------------------------------------------------
 * ---------------------------------------------------------------------------
 */
package org.ow2.petals.se.eip.async;

import org.ow2.petals.component.framework.api.message.Exchange;

/**
 * @author Frederic Gardes
 * 
 */
public class CommonAsyncContext extends AbstractAsyncContext {

    private int ongoingSubMessages;

    private boolean faultRobust = false;

    private boolean exceptionRobust = false;

    /**
     * The Constructor
     * 
     * @param originalExchange
     * @param timeToLive
     * @param pattern
     * @param faultRobust
     * @param exceptionRobust
     * @param ongoingSubMessages
     */
    public CommonAsyncContext(final Exchange originalExchange, final Long timeToLive,
            final String pattern, final boolean faultRobust, final boolean exceptionRobust,
            final int ongoingSubMessages) {
        super(originalExchange, timeToLive, pattern);
        this.faultRobust = faultRobust;
        this.exceptionRobust = exceptionRobust;
        this.ongoingSubMessages = ongoingSubMessages;
    }

    /**
     * Get the faultRobust
     * 
     * @return the faultRobust
     */
    public boolean isFaultRobust() {
        return this.faultRobust;
    }

    /**
     * Get the exceptionRobust
     * 
     * @return the exceptionRobust
     */
    public boolean isExceptionRobust() {
        return this.exceptionRobust;
    }

    /**
     * Decrement and get the number of messages to wait
     * 
     * @return The number of messages to wait
     */
    public int decrementAndGetOngoingSubMessages() {
        this.ongoingSubMessages--;
        return this.ongoingSubMessages;
    }
}
