/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.SortedMap;
import java.util.TreeMap;

import javax.jbi.JBIException;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.xpath.XPathExpressionException;

import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.ow2.petals.component.framework.util.XMLUtil;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;
import org.w3c.dom.Document;

/**
 * A Router pattern
 * 
 * @author alouis
 * @author ofabre
 * @author Christophe HAMERLING - eBM WebSourcing
 * @author Frederic Gardes
 * @since 2.2
 * 
 */
public class DynamicRouter extends AbstractPattern {

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.util.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public boolean process(final Exchange exchange, final ExchangeContext context) {
        try {
            final List<Consumes> consumesList = context.getSUConsumes(exchange.getEndpoint());
            final List<String> conditions = this.retrieveConditions(context.getExtensions());
            // there must be numberOfConditions+2 consumes defined in the SU
            if (consumesList != null) {
                final int count = consumesList.size();
                if (count == conditions.size() + 2) {
                    // Reset the In content as a Document Source, to avoid the
                    // Streaming
                    // issue
                    final Document inMessage = exchange.getInMessageContentAsDocument();
                    exchange.getInMessage().setContent(new DOMSource(inMessage));
                    // 1. the first' consumes' is the targeted service used to
                    // evaluate the
                    // condition
                    final Consumes conditionService = consumesList.get(0);
                    if (conditionService != null) {
                        final Exchange conditionExchange = context
                        .createConsumeExchange(conditionService);
                        PatternHelper.copy(exchange.getInMessage(), conditionExchange
                                .getInMessage());
                        UtilFactory.getExchangeUtil().copyExchangeProperties(exchange,
                                conditionExchange);
                        if (conditionExchange.getOperation() == null) {
                            conditionExchange.setOperation(exchange.getOperation());
                        }
                        if (context.sendSync(conditionExchange)) {
                            // 2. process the result of the Service to evaluate
                            // find the Consumes to call depending on the xpath
                            // conditions
                            NormalizedMessage receivedConditionMessage = null;
                            receivedConditionMessage = PatternHelper.processSentExchange(
                                    conditionExchange, context);

                            if (receivedConditionMessage != null) {
                                if (!(receivedConditionMessage instanceof Fault)) {
                                    final Document conditionMessage = UtilFactory.getSourceUtil()
                                    .createDocument(receivedConditionMessage.getContent());
                                    // 2.1 evaluate the result with conditions
                                    // to retrieve the
                                    // Consumes
                                    // to call
                                    // (default Consumes is used if no
                                    // conditions is True)
                                    int targetedConsumesListIndentifier;
                                    targetedConsumesListIndentifier = this
                                    .retrieveTargetedConsumesListIdentifier(conditions,
                                            conditionMessage, context);
                                    // 2.2 forward the original exchange to the
                                    // Service
                                    final Consumes consumedService = consumesList
                                    .get(targetedConsumesListIndentifier);
                                    if (consumedService.getMep() != null) {
                                        final Exchange routedExchange = context
                                        .createConsumeExchange(consumedService);
                                        PatternHelper.copy(exchange.getInMessage(), routedExchange
                                                .getInMessage());
                                        UtilFactory.getExchangeUtil().copyExchangeProperties(
                                                exchange, routedExchange);
                                        if (routedExchange.getOperation() == null) {
                                            routedExchange.setOperation(exchange.getOperation());
                                        }
                                        this.logSend(consumedService, routedExchange);
                                        if (context.sendSync(routedExchange)) {
                                            // 2.3 process the response
                                            NormalizedMessage receivedMessage = null;
                                            receivedMessage = PatternHelper.processSentExchange(
                                                    routedExchange, context);
                                            // if a response has been returned,
                                            // set it on the
                                            // original
                                            // exchange
                                            // otherwise, the'
                                            // processSentExchange()' has manage
                                            // the
                                            // exchange
                                            if (receivedMessage != null) {
                                                if (receivedMessage instanceof Fault) {
                                                    if (!exchange.isInOnlyPattern()) {
                                                        exchange.setFault((Fault) receivedMessage);
                                                    }
                                                } else {
                                                    if (!exchange.isInOnlyPattern()
                                                            && !exchange.isRobustInOnlyPattern()) {
                                                        exchange.setOutMessage(receivedMessage);
                                                    }
                                                }
                                            } else {
                                                exchange
                                                .setError(new JBIException(
                                                        "Timeout occurred, no reponse has been received from the routed service"));
                                            }
                                        } else {
                                            final StringBuffer message = new StringBuffer(
                                            "Unable to found the Message Exchange Pattern to use to send the exchange to");
                                            boolean first = true;
                                            if (consumedService.getInterfaceName() != null
                                                    && consumedService.getInterfaceName()
                                                    .getNamespaceURI() != null
                                                    && consumedService.getInterfaceName()
                                                    .getLocalPart() != null) {
                                                first = false;
                                                message.append(" the interface "
                                                        + consumedService.getInterfaceName()
                                                        .toString());
                                            }
                                            if (consumedService.getServiceName() != null
                                                    && consumedService.getServiceName()
                                                    .getNamespaceURI() != null
                                                    && consumedService.getServiceName()
                                                    .getLocalPart() != null) {
                                                if (!first) {
                                                    message.append(",");
                                                }
                                                first = false;
                                                message.append(" the service "
                                                        + consumedService.getInterfaceName()
                                                        .toString());
                                            }
                                            if (consumedService.getEndpointName() != null) {
                                                if (!first) {
                                                    message.append(" and");
                                                }
                                                first = false;
                                                message.append(" the endpoint name "
                                                        + consumedService.getEndpointName());
                                            }
                                            if (first) {
                                                message.append(" an unknow service");
                                            }
                                            message
                                            .append(". Please add it to the description of your service, into the Consumes part");
                                            exchange.setError(new Exception(message.toString()));
                                        }
                                    } else {
                                        exchange.setError(new Exception(this.getName()
                                                + ": the test service return a fault, "
                                                + UtilFactory.getSourceUtil().createString(
                                                        receivedConditionMessage.getContent())));
                                    }
                                }
                            } else {
                                exchange
                                .setError(new JBIException(
                                        this.getName()
                                        + ": timeout occurred, no reponse has been received from the evaluation service"));
                            }
                        }
                    } else {
                        exchange.setError(new JBIException(this.getName()
                                + ": the Service Unit must define a valid condition service"));
                    }
                } else {
                    exchange
                    .setError(new JBIException(
                            this.getName()
                            + ": the Service Unit must define (numberOfConditions+2) Consumes sections for the pattern"));
                }
            } else {
                exchange
                .setError(new JBIException(
                        this.getName()
                        + ": the Service Unit must define (numberOfConditions+2) Consumes sections for the pattern"));
            }
        } catch (final JBIException e) {
            exchange.setError(new Exception(e));
        } catch (final XPathExpressionException e) {
            exchange.setError(new Exception(e));
        }
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#getName()
     */
    public String getName() {
        return "Dynamic-Router";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    public boolean validateMEP(final URI mep) {
        return MEPConstants.IN_ONLY_PATTERN.equals(mep)
        || MEPConstants.ROBUST_IN_ONLY_PATTERN.equals(mep)
        || MEPConstants.IN_OUT_PATTERN.equals(mep);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /**
     * 
     * @param conditions
     * @param inMessage
     * @param context
     * @return
     * @throws XPathExpressionException
     */
    private int retrieveTargetedConsumesListIdentifier(final List<String> conditions,
            final Document inMessage, final ExchangeContext context)
    throws XPathExpressionException {
        // Choose the default one first
        int targetedConsumesListIndentifier = conditions.size() + 1;
        for (int i = 0; i < conditions.size(); i++) {
            // evaluate the IN message
            try {
                final String evaluation = XMLUtil.createStringFromDOMDocument(inMessage);
                this.logger.fine(this.getName() + ": document to evaluate : " + evaluation);
            } catch (final TransformerException e) {
                this.logger.warning(this.getName() + ": wrong document to evaluate");
            }
            this.logger.fine(this.getName() + ": evaluate condition #" + (i + 1) + " : "
                    + conditions.get(i));
            if (PatternHelper.evaluateXPathCondition(conditions.get(i), inMessage.getFirstChild())) {
                this.logger.fine(this.getName() + ": evaluation is OK");
                targetedConsumesListIndentifier = i + 1;
                break;
            }
        }
        if (targetedConsumesListIndentifier == conditions.size() + 1) {
            this.logger.fine(this.getName()
                    + ": no condition is satisfied, default Service will be called");
        }
        return targetedConsumesListIndentifier;
    }

    /**
     * 
     * @param extensions
     * @return
     */
    private List<String> retrieveConditions(final ConfigurationExtensions extensions) {
        final List<String> conditions = new ArrayList<String>();
        final SortedMap<Integer, String> orderedConditions = new TreeMap<Integer, String>();

        for (final String param : extensions.keySet()) {
            if (EIPConstants.TEST.equals(param)) {
                conditions.add(extensions.get(param));
                break;
            } else if (param.startsWith(EIPConstants.TEST)) {
                final String indexString = param.substring(EIPConstants.TEST.length());
                orderedConditions.put(Integer.valueOf(indexString), extensions.get(param));
            }
        }

        if (conditions.size() == 0) {
            conditions.addAll(orderedConditions.values());
        }

        return conditions;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#processAsync(org.ow2.petals
     * .component.framework.api.message.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext,
     * org.ow2.petals.se.eip.async.CommonAsyncContext)
     */
    @Override
    public boolean processAsync(final Exchange exchange, final ExchangeContext context,
            final CommonAsyncContext asyncContext) {
        exchange.setError(new Exception(this.getName()
                + ": the pattern doesn't process the asynchronous exchanges"));
        return true;
    }
}
