/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.dom.DOMSource;
import javax.xml.xpath.XPathExpressionException;

import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;
import org.w3c.dom.Document;

/**
 * A Router pattern
 * 
 * @author alouis
 * @author ofabre
 * @author Christophe HAMERLING - eBM WebSourcing
 * @since 2.2
 * 
 */
public class Router extends AbstractPattern {

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.util.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public boolean process(final Exchange exchange, final ExchangeContext context) {
        try {
            final List<Consumes> consumesList = context.getSUConsumes(exchange.getEndpoint());
            final List<String> conditions = this.retrieveConditions(context.getExtensions());
            // there must be numberOfConditions+1' consumes' defined in the SU
            if (consumesList != null) {
                final int count = consumesList.size();
                if (count == conditions.size() + 1) {
                    // reset the In content as a Document Source, to avoid the
                    // Streaming
                    // issue
                    final Document document = UtilFactory.getSourceUtil().createDocument(
                            exchange.getInMessage().getContent());
                    exchange.getInMessage().setContent(new DOMSource(document));
                    // retrieve identifier of the first consumes linked to a
                    // true
                    // condition, or the default one
                    int targetedConsumesListIndentifier;
                    targetedConsumesListIndentifier = this.retrieveTargetedConsumesListIdentifier(
                            conditions, document, context);
                    // forward the exchange to the correct service, depending on
                    // the
                    // test result
                    final Consumes consumedService = consumesList
                    .get(targetedConsumesListIndentifier);
                    final Exchange routedExchange = context.createConsumeExchange(consumedService);
                    PatternHelper.copy(exchange.getInMessage(), routedExchange.getInMessage());
                    UtilFactory.getExchangeUtil().copyExchangeProperties(exchange, routedExchange);
                    if (routedExchange.getOperation() == null) {
                        routedExchange.setOperation(exchange.getOperation());
                    }
                    this.logSend(consumedService, routedExchange);
                    if (context.sendSync(routedExchange)) {
                        if (this.logger == null) {
                            this.logger = context.getLogger();
                            if (this.logger == null) {
                                this.logger = Logger.getLogger(AbstractPattern.class.getName());
                            }
                        }
                        // process the response
                        NormalizedMessage receivedMessage = null;
                        receivedMessage = PatternHelper
                        .processSentExchange(routedExchange, context);
                        // if a response has been returned, set it on the
                        // original exchange.
                        // otherwise, the processSentExchange() has managed the
                        // exchange
                        if (receivedMessage != null) {
                            if (receivedMessage instanceof Fault) {
                                PatternHelper.copy(receivedMessage, exchange.getFault());
                            } else {
                                PatternHelper.copy(receivedMessage, exchange.getOutMessage());
                            }
                        }
                    } else if (exchange.isActiveStatus()) {
                        exchange
                        .setError(new Exception(
                                "Timeout occurred, no reponse has been received from the target service"));
                    }
                } else {
                    exchange
                    .setError(new JBIException(
                            this.getName()
                            + ": the Service Unit must define (numberOfConditions+1) Consumes sections for the pattern "));
                }
            } else {
                exchange
                .setError(new JBIException(
                        this.getName()
                        + ": the Service Unit must define (numberOfConditions+2) Consumes sections for the pattern "));
            }

        } catch (final JBIException e) {
            exchange.setError(new Exception(e));
        } catch (final XPathExpressionException e) {
            exchange.setError(new Exception(e));
        }
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#getName()
     */
    public String getName() {
        return "Router";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    public boolean validateMEP(final URI mep) {
        return MEPConstants.IN_ONLY_PATTERN.equals(mep)
        || MEPConstants.ROBUST_IN_ONLY_PATTERN.equals(mep)
        || MEPConstants.IN_OUT_PATTERN.equals(mep);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /**
     * 
     * @param conditions
     * @param inMessage
     * @param context
     * @return
     * @throws XPathExpressionException
     */
    private int retrieveTargetedConsumesListIdentifier(final List<String> conditions,
            final Document inMessage, final ExchangeContext context)
    throws XPathExpressionException {
        int targetedConsumesListIndentifier = conditions.size();
        for (int i = 0; i < conditions.size(); i++) {
            // evaluate the IN message
            this.logger.fine(this.getName() + ": evaluates condition #" + (i + 1) + " : "
                    + conditions.get(i));
            if (conditions.get(i) != null
                    && inMessage.getDocumentElement() != null
                    && PatternHelper.evaluateXPathCondition(conditions.get(i), inMessage
                            .getDocumentElement())) {
                targetedConsumesListIndentifier = i;
                break;
            }
        }
        return targetedConsumesListIndentifier;
    }

    /**
     * 
     * @param extensions
     * @return
     */
    private List<String> retrieveConditions(final ConfigurationExtensions extensions) {
        final List<String> conditions = new ArrayList<String>();
        final SortedMap<Integer, String> orderedConditions = new TreeMap<Integer, String>();

        for (final String param : extensions.keySet()) {
            if (EIPConstants.TEST.equals(param)) {
                conditions.add(extensions.get(param));
                break;
            } else if (param.startsWith(EIPConstants.TEST)) {
                final String indexString = param.substring(EIPConstants.TEST.length());
                orderedConditions.put(Integer.valueOf(indexString), extensions.get(param));
            }
        }

        if (conditions.size() == 0) {
            conditions.addAll(orderedConditions.values());
        }

        return conditions;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#processAsync(org.ow2.petals
     * .component.framework.api.message.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext,
     * org.ow2.petals.se.eip.async.CommonAsyncContext)
     */
    @Override
    public boolean processAsync(final Exchange exchange, final ExchangeContext context,
            final CommonAsyncContext asyncContext) {
        exchange.setError(new Exception(this.getName()
                + ": the pattern doesn't process the asynchronous exchanges"));
        return true;
    }
}
